@@ -1,4 +1,5 @@
 from datetime import timedelta
+from expvar.stats import stats
 import logging
 from oid_translate import ObjectId
 
@@ -10,11 +11,20 @@ from pysnmp.proto.error import ProtocolError
 import socket
 from sqlalchemy.exc import IntegrityError, InvalidRequestError, OperationalError
 
+from trapperkeeper.dde import DdeNotification
 from trapperkeeper.constants import SNMP_VERSIONS
 from trapperkeeper.models import Notification
 from trapperkeeper.utils import parse_time_string, send_trap_email
 
 
+try:
+    from trapperkeeper_dde_plugin import run as dde_run
+except ImportError as err:
+    # If no DDE plugin is found add noop to namespace.
+    def dde_run(notification):
+        pass
+
+
 class TrapperCallback(object):
     def __init__(self, conn, template_env, config, resolver):
         self.conn = conn
@@ -29,6 +39,7 @@ class TrapperCallback(object):
         # Prevent the application from crashing when callback raises
         # an exception.
         except Exception as err:
+            stats.incr("callback-failure", 1)
             logging.exception("Callback Failed: %s", err)
 
     def _send_mail(self, handler, trap, is_duplicate):
@@ -51,9 +62,12 @@ class TrapperCallback(object):
         }
         ctxt = dict(trap=trap, dest_host=self.hostname)
         try:
+            stats.incr("mail_sent_attempted", 1)
             send_trap_email(recipients, "trapperkeeper",
                             subject, self.template_env, ctxt)
+            stats.incr("mail_sent_successful", 1)
         except socket.error as err:
+            stats.incr("mail_sent_failed", 1)
             logging.warning("Failed to send e-mail for trap: %s", err)
 
     def _call(self, transport_dispatcher, transport_domain, transport_address, whole_msg):
@@ -65,6 +79,7 @@ class TrapperCallback(object):
         if msg_version in api.protoModules:
             proto_module = api.protoModules[msg_version]
         else:
+            stats.incr("unsupported-notification", 1)
             logging.error("Unsupported SNMP version %s", msg_version)
             return
 
@@ -74,6 +89,7 @@ class TrapperCallback(object):
         try:
             req_msg, whole_msg = decoder.decode(whole_msg, asn1Spec=proto_module.Message(),)
         except (ProtocolError, ValueConstraintError) as err:
+            stats.incr("unsupported-notification", 1)
             logging.warning("Failed to receive trap (%s) from %s: %s", version, host, err)
             return
         req_pdu = proto_module.apiMessage.getPDU(req_msg)
@@ -82,15 +98,25 @@ class TrapperCallback(object):
         # community = proto_module.apiMessage.getCommunity(req_msg)
 
         if not req_pdu.isSameTypeWith(proto_module.TrapPDU()):
+            stats.incr("unsupported-notification", 1)
             logging.warning("Received non-trap notification from %s", host)
             return
 
         if msg_version not in (api.protoVersion1, api.protoVersion2c):
+            stats.incr("unsupported-notification", 1)
             logging.warning("Received trap not in v1 or v2c")
             return
 
         trap = Notification.from_pdu(host, proto_module, version, req_pdu)
-        handler = self.config.handlers[trap.oid]
+        if trap is None:
+            stats.incr("unsupported-notification", 1)
+            logging.warning("Invalid trap from %s: %s", host, req_pdu)
+            return
+
+        dde = DdeNotification(trap, self.config.handlers[trap.oid])
+        dde_run(dde)
+        handler = dde.handler
+
         trap.severity = handler["severity"]
         trap.manager = self.hostname
 
@@ -99,31 +125,35 @@ class TrapperCallback(object):
             expires = timedelta(**expires)
             trap.expires = trap.sent + expires
 
-        if trap is None:
-            logging.warning("Invalid trap from %s: %s", host, req_pdu)
-            return
-
+        stats.incr("traps_received", 1)
         objid = ObjectId(trap.oid)
         if handler.get("blackhole", False):
+            stats.incr("traps_blackholed", 1)
             logging.debug("Blackholed %s from %s", objid.name, host)
             return
 
         logging.info("Trap Received (%s) from %s", objid.name, host)
+        stats.incr("traps_accepted", 1)
 
 
         duplicate = False
         try:
+            stats.incr("db_write_attempted", 1)
             self.conn.add(trap)
             self.conn.commit()
+            stats.incr("db_write_successful", 1)
         except OperationalError as err:
             self.conn.rollback()
             logging.warning("Failed to commit: %s", err)
+            stats.incr("db_write_failed", 1)
             # TODO(gary) reread config and reconnect to database
         except InvalidRequestError as err:
             # If we get into this state we should rollback any pending changes.
+            stats.incr("db_write_failed", 1)
             self.conn.rollback()
             logging.warning("Bad state, rolling back transaction: %s", err)
         except IntegrityError as err:
+            stats.incr("db_write_duplicate", 1)
             duplicate = True
             self.conn.rollback()
             logging.info("Duplicate Trap (%s) from %s. Likely inserted by another manager.", objid.name, host)
